﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using ClientServer.Controllers.Core.Testing;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Models.Files;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Primitives;
using Newtonsoft.Json.Linq;

namespace ClientServer.Controllers.Core.Exercises
{
    [Route(Constants.ApiPrefix + "exercise/do")]
    public class DoExerciseController : ControllerWithDb
    {
        /*
         * after accessing the exercise (get) the user has a participation (ExerciseReleaseWithUserAsParticipation), this is ensured by the (get) method
         */


        public DoExerciseController(SyndromDbContext context) : base(context)
        {
        }


        /// <summary>
        /// returns the exercise for the given release (by code)
        /// this includes the custom tests for the exercise created by the user
        /// TODO get requests changes state... it created the participation if not existing
        /// </summary>
        /// <param name="generatedCode">the code for the exercise release</param>
        [HttpGet("{generatedCode}")]
        public async Task GetExercise(string generatedCode)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (string.IsNullOrWhiteSpace(generatedCode))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "release code was empty")));
                return;
            }


            int lastEditedPLangId = -1;

            //TODO MAYBE make performance better??
            ExerciseRelease exerciseRelease = await _context.ExerciseReleases
                    .Where(p => p.GeneratedCode == generatedCode
                                &&
                                (p.IsReleased ||
                                 p.ExerciseReleaseWithUserAsParticipations.Count(parti => parti.UserId == userId) == 1))
                    //is released or the user has a participation
                    .Include(p => p.Exercise)
                    .Include(p => p.PLang)

                    //test cases
                    .Include(p => p.Exercise.Tests)
                    .ThenInclude(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                    .Include(p => p.Exercise.Tests)
                    .ThenInclude(p => p.TestSettings)
                    .Include(p => p.Exercise.Tests)
                    .ThenInclude(p => p.TestType)

                    //code template
                    .Include(p => p.Exercise.CodeTemplates)
                    .Include(p => p.Exercise.CodeTemplates)
                    .ThenInclude(p => p.TemplateFiles)
                    .Include(p => p.Exercise.CodeTemplates)
                    .ThenInclude(p => p.MainFile)
                    .Include(p => p.Exercise.CodeTemplates)
                    .ThenInclude(p => p.PLang)

                    //task description
                    .Include(p => p.Exercise.Description)
                    .ThenInclude(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceMarkdownAsset)
                    .FirstOrDefaultAsync()
                ;


            if (exerciseRelease?.Exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise or release not found")));
                return;
            }

            Exercise exercise = exerciseRelease.Exercise;

            if (exercise.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exercise.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            ExerciseReleaseWithUserAsParticipation oldParticipation = await _context
                    .ExerciseReleaseWithUserAsParticipations
                    .Include(p => p.CustomTests)
                    .ThenInclude(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceUserFileAsset)
                    .Include(p => p.CustomTests)
                    .ThenInclude(p => p.TestResultsNew)
                    .FirstOrDefaultAsync(p => p.ExerciseReleaseId == exerciseRelease.Id && p.UserId == userId)
                ;

            if (oldParticipation == null && exerciseRelease.IsReleased == false)
            {
                //user has no old participation and the release is not released --> no access
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }


            #region forge exercise for frontend

            var filesIdsToRead = exercise.Description.AssetReferences
                .Select(p => p.FileReferenceMarkdownAssetId)
                .ToArray();

            var basePath = Files.GetUploadFilePath(UploadDirType.MarkdownAssets);

            var filesDict = await Files.ReadUploadedFiles(basePath, filesIdsToRead);

            var exerciseForFrontend = new DoExerciseFromBackend
            {
                Id = exercise.Id,
                CanUserCreateFiles = exercise.CanUserCreateFiles,
                CanUserCreateCustomTests = exercise.CanUserCreateCustomTests,
                IsPermanentlyLocked = exercise.IsPermanentlyLocked,
                DisplayName = exercise.DisplayName,
                ShortDescription = exercise.ShortDescription,
                ExerciseRelease = new ExerciseReleaseDoExerciseFromBackend
                {
                    Id = exerciseRelease.Id,
                    AvailableWorkingTimeInMinutes = exerciseRelease.AvailableWorkingTimeInMinutes,
                    HasLimitedWorkingTime = exerciseRelease.HasLimitedWorkingTime,
                    PLangId = exerciseRelease.PLangId,
                    ReleaseCode = exerciseRelease.GeneratedCode,
                    IsReleased = exerciseRelease.IsReleased,
                    AutomaticEndAt =
                        exerciseRelease.ReleaseDurationType == ReleaseDurationType.Infinitely
                            ? null
                            : exerciseRelease.AutomaticEndAt,
                    HasAutomaticAssessmentFinished = exerciseRelease.HasAutomaticAssessmentFinished,
                    ReleaseDurationType = (int) exerciseRelease.ReleaseDurationType,
                    ReleaseStartType = (int) exerciseRelease.ReleaseStartType,
                    ShouldClearClipboard = exerciseRelease.ShouldClearClipboard,
                    HideSiteHeaderBar = exerciseRelease.HideSiteHeaderBar,
                    HideExerciseLeaveActions = exerciseRelease.HideExerciseLeaveActions,
                    HidePrintOptions = exerciseRelease.HidePrintOptions,
                    ShowAdditionalLogButton = exerciseRelease.ShowAdditionalLogButton,
                    MaxManualPoints = exerciseRelease.MaxManualPoint,
                },
                ExerciseDescription = new ExerciseDescriptionFullBase
                {
                    Id = exercise.Description.Id,
                    Content = exercise.Description.Content,
                    Assets =
                        exercise.Description.AssetReferences.Select(
                            asset => new FileWithData
                            {
                                Id = asset.FileReferenceMarkdownAsset.Id,
                                DisplayName = asset.FileReferenceMarkdownAsset.OriginalName,
                                MimeType = asset.FileReferenceMarkdownAsset.MimeType,
                                Content = filesDict[asset.FileReferenceMarkdownAssetId]
                            }).ToList()
                },
                LockSolutionsFlag = oldParticipation == null ? false : oldParticipation.LockSolutionsFlag
            };

            #endregion

            #region code templates

            var codeTemplatesToUse = new List<CodeTemplate>();

            if (exerciseRelease.PLang == null)
            {
                //include all languages...
                codeTemplatesToUse = exercise.CodeTemplates;

                //a exercise needs at least 1 plang tempalte.... that is ensured
            }
            else
            {
                //only include languages for the specific release with the specific p language
                var codeTemplate = exercise.CodeTemplates.FirstOrDefault(p => p.PLangId == exerciseRelease.PLang.Id);


                //if user specified a plang for the release but there is no template for this plang in the exercise
                //then return error
                if (codeTemplate == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no template for plang found")));
                    return;
                }

                codeTemplatesToUse.Add(codeTemplate);
            }

            //add all code templates to the exercise for the frontend
            foreach (var codeTemplate in codeTemplatesToUse)
            {
                if (codeTemplate.MainFile == null)
                {
                    Console.WriteLine("WARNING a code template has no main file...this sould not happen!");
                    continue;
                }

                var template = new ExerciseCodeTemplateFullBase()
                {
                    Id = codeTemplate.Id,
                    PLangId = codeTemplate.PLangId,
                    MainFileId = codeTemplate.MainFile.Id,
                    TemplateFiles = codeTemplate.TemplateFiles.Select(temp => new ExerciseCodeTemplateFileFullBase()
                    {
                        Id = temp.Id,
                        Content = temp.IsContentVisibleForUser ? temp.Content : "",
                        IsEditableByUser = temp.IsEditableByUser,
                        FileNameWithExtension = temp.FileNameWithExtension,
                        IsContentVisibleForUser = temp.IsContentVisibleForUser
                    }).ToList()
                };


                if (codeTemplatesToUse.Count == 0)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "no templates found")));
                    return;
                }


                if (oldParticipation?.LastEditedPLangId != null)
                {
                    //pre select the last plang the user edited...
                    lastEditedPLangId = oldParticipation.LastEditedPLangId;
                }
                else
                {
                    //if plang set and code template for this plang exists
                    if (codeTemplatesToUse.Any(p => p.PLangId == exerciseRelease.PLangId))
                    {
                        lastEditedPLangId = exerciseRelease.PLangId;
                    }
                    else
                    {
                        //first time the user do the exercise to just take the first code template
                        lastEditedPLangId = codeTemplatesToUse[0].PLangId;
                    }
                }


                exerciseForFrontend.CodeTemplates.Add(template);
            }

            exerciseForFrontend.LastEditedPLangId = lastEditedPLangId;

            #endregion

            #region tests

            foreach (var test in exercise.Tests)
            {
                if (test.IsSubmitTest == false)
                {
                    exerciseForFrontend.Tests.Add(new ExerciseTestFromBackend
                    {
                        Id = test.Id,
                        DisplayName = test.DisplayName,
                        Content = test.Content,
                        Weight = test.Weight,
                        TestTypeId = test.TestType.Id,
                        TestSettings = new TestSettingsFullBase
                        {
                            Id = test.TestSettings.Id,
                            MemoryLimitInKb = test.TestSettings.MemoryLimitInKb,
                            TimeoutInMs = test.TestSettings.TimeoutInMs,
                            MaxDiskSpaceInKb = test.TestSettings.MaxDiskSpaceInKb,
                            CompileTimeoutInMs = test.TestSettings.CompileTimeoutInMs,
                            CompilerOptions = test.TestSettings.CompilerOptions,
                        },
                        Files = test.AssetReferences.Select(p => new FilePreviewFromBackend()
                        {
                            Id = p.FileReferenceTestAsset.Id,
                            OriginalName = p.FileReferenceTestAsset.OriginalName,
                            MimeType = p.FileReferenceTestAsset.MimeType,
                            SizeInBytes = p.FileReferenceTestAsset.SizeInBytes,
                        }).ToList(),
                        IsSubmitTest = false, //we never give the student the submit tests
                        DisplayIndex = test.DisplayIndex
                    });
                }
            }

            #endregion

            //get custom tests if any

            #region (default) custom test settings

            //get DefaultCustomTestCaseSettings
            DefaultCustomTestSettings defaultCustomTestSettings = await _context.DefaultCustomTestSettings
                .FirstOrDefaultAsync(p => p.ExerciseId == exercise.Id);

            if (defaultCustomTestSettings == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no custom test settings found")));
                return;
            }

            exerciseForFrontend.DefaultCustomTestSettings = new TestSettingsFullBase()
            {
                Id = defaultCustomTestSettings.Id,
                MaxDiskSpaceInKb = defaultCustomTestSettings.MaxDiskSpaceInKb,
                TimeoutInMs = defaultCustomTestSettings.TimeoutInMs,
                MemoryLimitInKb = defaultCustomTestSettings.MemoryLimitInKb,
                CompileTimeoutInMs = defaultCustomTestSettings.CompileTimeoutInMs,
                CompilerOptions = defaultCustomTestSettings.CompilerOptions
            };

            #endregion

            #region custom tests

            if (oldParticipation != null)
            {
                exerciseForFrontend.CustomTests = new List<CustomTestFromBackend>();


                if (exerciseForFrontend.CanUserCreateCustomTests == true)
                {
                    foreach (CustomTest oldCustomTest in oldParticipation.CustomTests)
                    {
                        var testForBackend = new CustomTestFromBackend()
                        {
                            Id = oldCustomTest.Id,
                            DisplayName = oldCustomTest.DisplayName,
                            Content = oldCustomTest.Content,
                            TestTypeId = oldCustomTest.TestTypeId,
                            DisplayIndex = oldCustomTest.DisplayIndex,
                            Files = oldCustomTest.AssetReferences.Select(asset => new FilePreviewFromBackend()
                            {
                                Id = asset.FileReferenceUserFileAsset.Id,
                                OriginalName = asset.FileReferenceUserFileAsset.OriginalName,
                                MimeType = asset.FileReferenceUserFileAsset.MimeType,
                                SizeInBytes = asset.FileReferenceUserFileAsset.SizeInBytes,
                            }).ToList(),
                        };

                        //do not include custom test results here because they are plang dependent (include when getting user solution)

                        exerciseForFrontend.CustomTests.Add(testForBackend);
                    }
                }
            }

            #endregion

            //the exercise is visible to all but the student does not have a participation...
            //the students needs a participation in order to solve the exercise so create one...
            if (exerciseRelease.IsVisibleToAllAfterRelease && oldParticipation == null)
            {
                //no new participations allowed
                if (exercise.IsPermanentlyLocked)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is permanently locked")));
                    return;
                }

                int lastEditedPlangId = -1;

                if (codeTemplatesToUse.Count == 0)
                {
                    //should never happen...
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                "no code templates found for exercise")));
                    return;
                }


                lastEditedPlangId = exerciseRelease.PLangId;

                oldParticipation = new ExerciseReleaseWithUserAsParticipation
                {
                    ExerciseRelease = exerciseRelease,
                    UserId = userId,
                    //must be set here (in the model its never 0)
                    LastEditedPLangId = lastEditedPlangId,
                    LockSolutionsFlag = false
                };

                try
                {
                    _context.ExerciseReleaseWithUserAsParticipations.Add(oldParticipation);
                    await _context.SaveChangesAsync();
                }
                catch (Exception ex)
                {
                    await base.HandleDbError(ex);
                    return;
                }
            }
            else if (oldParticipation == null) //can this really happen?
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "participation not found")));
                return;
            }

            //ensures that the user has a participation after getting the exercise
            //because the user has entered a code --> participation created or released to all and now accessing (see above)
            //so normally this is always ensured
            Contract.Ensures(oldParticipation != null);
            //same but more clearer
            //Contract.Ensures(_context.ExerciseReleaseWithUserAsParticipations
            //.FirstOrDefault(p => p.ExerciseReleaseId == exerciseRelease.Id && p.UserId == userId) != null);


            //after we have an old participation (is ensured here) we can set the left working time...
            //when the participation is inserted to the db in this method then the +2gtm (timezone) is not stored to the db but the date is created here?
            //so when receiving the date from teh db the zimezone is missing but we do not receive the date time from the db because we just inserted and 
            //not other query run --> we still have the timezone
            exerciseForFrontend.FirstAccessedAt = DateTimeHelper.SetKind(oldParticipation.CreatedAt);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<DoExerciseFromBackend>(ResponseCode.Ok,
                        "do exercise queried",
                        exerciseForFrontend)));
        }

        /// <summary>
        /// sets the users solution should not count to true (can only be set true once and never back (not by the student))
        /// </summary>
        [HttpPost("solution/shouldnotcount/{generatedCode}")]
        public async Task SetShouldNotCount(string generatedCode)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            ExerciseReleaseWithUserAsParticipation releaseWithUserAsParticipation = await _context
                    .ExerciseReleaseWithUserAsParticipations
                    .Where(p => p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode)
                    .Include(p => p.ExerciseRelease.Exercise)
                    .FirstOrDefaultAsync()
                ;


            if (releaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "participation not found")));
                return;
            }


            if (releaseWithUserAsParticipation?.ExerciseRelease == null)
            {
                //this should not happen because when accessing the exercise (getting exercise) then the participation is ensured...

                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            if (releaseWithUserAsParticipation?.ExerciseRelease?.Exercise == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }


            if (releaseWithUserAsParticipation.ExerciseRelease.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "exercise is permanently locked")));
                return;
            }

            if (releaseWithUserAsParticipation.ExerciseRelease.Exercise.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != releaseWithUserAsParticipation.ExerciseRelease.Exercise.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            //check if available working time is up
            if (ReleaseHelper.IsTimeUp(releaseWithUserAsParticipation,
                releaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }

            releaseWithUserAsParticipation.ShouldNotCount = true;
            releaseWithUserAsParticipation.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        /// <summary>
        /// returns the user solution for the given release (exercise) and plang
        ///
        /// do not check if only visible to owner because this might be need in other controllers
        ///   this does not modify the state...
        /// </summary>
        /// <param name="generatedCode">the exercise release</param>
        /// <param name="pLangId"></param>
        [HttpGet("solution/{generatedCode}/{pLangId}")]
        public async Task GetSolutionForExercise(string generatedCode, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            await _GetSolutionForExercise(generatedCode, pLangId, userId);
        }

        internal async Task _GetSolutionForExercise(string generatedCode, int pLangId, int targetUserId)
        {
            int userId = targetUserId;

            if (string.IsNullOrWhiteSpace(generatedCode))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise release not found")));
                return;
            }

            //ensure plang exists
            PLang pLang = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == pLangId);

            if (pLang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "plang not found")));
                return;
            }

            //now load the user solution (if any)

            ExerciseReleaseWithUserAsParticipation releaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .Include(p => p.ExerciseRelease.Exercise)
                .Include(p => p.ExerciseRelease.Exercise.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)
                .FirstOrDefaultAsync(p =>
                    p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode);


            if (releaseWithUserAsParticipation?.ExerciseRelease?.Exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "participation or release or exercise not found")));
                return;
            }


            ExerciseRelease release = releaseWithUserAsParticipation.ExerciseRelease;


            //now check if the user has a solution...
            //because when we test then we use the user solution and when the user first accesses an exercise
            //then there is no solution just the provided code templates
            //the first save would give us the user solution (the code templates if the user hadn't changed anything)
            //but the user could click the test button without having a solution...

            //to solve this create a solution from the code templates for the given plang
            //so after this method ends it must be ensured that the user has a solution for the given plang (and exercise)...

            Solution userSolution = await _context.Solutions
                .Include(p => p.PLang)
                .Include(p => p.SolutionFiles)
                .ThenInclude(p => p.TemplateFile)
                .Include(p => p.TestResults)
                .Include(p => p.CustomTestResults)
                .FirstOrDefaultAsync(
                    p =>
                        p.UserId == userId && p.ExerciseReleaseId == releaseWithUserAsParticipation.ExerciseReleaseId &&
                        p.PLangId == pLangId);


            if (userSolution == null)
            {
                if (release.Exercise.IsPermanentlyLocked)
                {
                    await
                        Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "exercise is permanently locked")));
                    return;
                }

                //user has no solution...create one from the code templates


                if (releaseWithUserAsParticipation.ExerciseRelease.IsReleased == false)
                {
                    //do not create a solution because the user did not created one in the time where the exercise was released..
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                                "no solution found")));
                    return;
                }


                //user has no solution and the release is active
                //BUT the user has a participation... (how can this happen??)
                if (ReleaseHelper.IsTimeUp(releaseWithUserAsParticipation,
                    releaseWithUserAsParticipation.ExerciseRelease))
                {
                    //if time is up then we no longer create an initial solution??
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                                "user has participation but no solution and time is up (but still released)...")));
                    return;
                }

                userSolution =
                    await _CreateInitialSolution(null, release, pLang, userId, releaseWithUserAsParticipation);


                if (userSolution == null) return;
            }


            //get templates for the p lang (to check if the code file is editable by the user or not)

            CodeTemplate template = releaseWithUserAsParticipation.ExerciseRelease.Exercise.CodeTemplates
                .FirstOrDefault(
                    p => p.PLangId == pLangId);

            if (template == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "code template not found")));
                return;
            }

            if (userSolution?.MainFileId == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "solution or main file not found")));
                return;
            }

            //template files should be loaded 
            var errorMsg = UserSolutionHelper.ReplaceReadonlyFiles(userSolution, template, true, false);

            if (errorMsg != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, errorMsg)));
                return;
            }

            //check if the user has an after solution

            bool hasLoadAfterSolution = await _context.AfterSolutions
                    .AnyAsync(p => p.SolutionUserId == userSolution.UserId &&
                                   p.SolutionExerciseReleaseId == userSolution.ExerciseReleaseId &&
                                   p.SolutionPLangId == userSolution.PLangId)
                ;

            bool shouldLoadAfterSolution = false;

            //check if available working time is up
            if (ReleaseHelper.IsTimeUp(releaseWithUserAsParticipation, releaseWithUserAsParticipation.ExerciseRelease))
            {
                //this is ok because now the user should see the after solution editor...
                shouldLoadAfterSolution = true;
            }
            else
            {
                shouldLoadAfterSolution = false;
            }

            var exerciseSolutionForFrontend = new SolutionDoExerciseFullBase()
            {
                UserId = userSolution.UserId,
                ReleaseId = userSolution.ExerciseReleaseId,
                MainFileId = userSolution.MainFileId.Value,
                PLangId = userSolution.PLangId,
                Note = userSolution.Note,
                SolutionFiles = new List<SolutionFileDoExerciseFullBase>(),
                LastUpdatedAt = userSolution.LastUpdatedAt,
                ShouldNotCount = releaseWithUserAsParticipation.ShouldNotCount,
                ShouldLoadAfterSolution = shouldLoadAfterSolution,
                HasLoadAfterSolution = hasLoadAfterSolution
            };

            //include all versions for the frontend
            var solutionVersionTupleFromBackend = new SolutionTupleFromBackend
            {
                Solution = exerciseSolutionForFrontend,
                FileIdAssociationMap = new List<FileIdAssociation>()
            };


            //now forge the solution files for frontend
            foreach (SolutionFile userSolutionPart in userSolution.SolutionFiles)
            {
                //use the user file content
                string fileContent = userSolutionPart.Content;
                string fileName = userSolutionPart.FileNameWithExtension;
                bool isEditableByUser = false;
                bool isContentVisibleForUser = false;
                TemplateFile oldTemplateFile = userSolutionPart.TemplateFile;

                if (oldTemplateFile == null)
                {
                    if (release.Exercise.CanUserCreateFiles)
                    {
                        //this is a tab the user has created... (no template file exists)
                        isEditableByUser = true; //the user created this file...
                    }
                    else
                    {
                        //a file without a template and user could not create files... (maybe the tutor changed the CanUserCreateFiles option??)
                        isEditableByUser = true;
                        //await Response.WriteAsync(Jc.Serialize(new BasicResponse(false, "file not found")));
                        //return;
                    }

                    isContentVisibleForUser = true;
                }
                else
                {
                    //reset to template code if readonly (to ensure same as template ...)
                    //this is already done above in UserSolutionHelper.ReplaceReadonlyFiles

                    isEditableByUser = oldTemplateFile.IsEditableByUser;
                    isContentVisibleForUser = oldTemplateFile.IsContentVisibleForUser;
                }


                exerciseSolutionForFrontend.SolutionFiles.Add(new SolutionFileDoExerciseFullBase()
                {
                    Id = userSolutionPart.Id,
                    Content = fileContent,
                    FileNameWithExtension = fileName,
                    DisplayIndex = userSolutionPart.DisplayIndex,
                    IsEditableByUser = isEditableByUser,
                    UiIsDisplayed = userSolutionPart.IsDisplayed,
                    TemplateFileId = userSolutionPart.TemplateFileId,
                    IsContentVisibleForUser = isContentVisibleForUser
                });
            }

            //this method must ensures that the user has a solution for this exercise (and release) for the plang
            //else the user could not run a test... see above
            Contract.Ensures(userSolution != null);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<SolutionTupleFromBackend>(ResponseCode.Ok,
                        "user solution queried",
                        solutionVersionTupleFromBackend)));
        }

        /// <summary>
        /// creates and saves a new user solution for the given release
        ///
        /// THIS DOES NOT check Iif there already exits a solution
        /// </summary>
        /// <param name="oldSolutionToDelete">old solution that should be deleted (for transaction) or null</param>
        /// <param name="release"></param>
        /// <param name="pLang"></param>
        /// <param name="userId"></param>
        /// <param name="releaseWithUserAsParticipation"></param>
        /// <returns>the created solution or null</returns>
        private async Task<Solution> _CreateInitialSolution(Solution oldSolutionToDelete, ExerciseRelease release,
            PLang pLang, int userId,
            ExerciseReleaseWithUserAsParticipation releaseWithUserAsParticipation)
        {
            Solution userSolution;

            #region create first user solution

            CodeTemplate codeTemplateForPLang =
                release.Exercise.CodeTemplates.FirstOrDefault(p => p.PLangId == pLang.Id);

            if (codeTemplateForPLang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "code template not found")));
                return null;
            }

            var solutionFiles = new List<SolutionFile>();
            SolutionFile mainFile = null;

            foreach (TemplateFile templateFile in codeTemplateForPLang.TemplateFiles)
            {
                var files = new SolutionFile()
                {
                    Content = templateFile.IsContentVisibleForUser ? templateFile.Content : "",
                    DisplayIndex = templateFile.DisplayIndex,
                    FileNameWithExtension = templateFile.FileNameWithExtension,
                    IsDisplayed = true,
                    TemplateFile = templateFile,
                };
                solutionFiles.Add(files);

                if (codeTemplateForPLang.MainFile == templateFile)
                {
                    mainFile = files;
                }
            }

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "main file not found")));
                return null;
            }


            userSolution = new Solution()
            {
                PLang = pLang,
                UserId = userId,
                ExerciseReleaseId = release.Id,
                ExerciseReleaseWithUserAsParticipation = releaseWithUserAsParticipation,
                SolutionFiles = solutionFiles,
                LastEditingIpAddress = NetworkHelper.GetIpAddress(HttpContext)
            };

            using (var transaction = _context.Database.BeginTransaction())
            {
                try
                {
                    if (oldSolutionToDelete != null)
                    {
                        //we need to unset the main file first else we get circular dependency error...
                        oldSolutionToDelete.MainFile = null;
                        await _context.SaveChangesAsync();
                        //this should automatically clear all old stuff (e.g. old test results)
                        _context.Solutions.Remove(oldSolutionToDelete);
                        await _context.SaveChangesAsync();
                    }

                    //first save solution
                    _context.Solutions.Add(userSolution);
                    await _context.SaveChangesAsync();

                    //then set main file else would give us circular dependency error on db update ...
                    userSolution.MainFile = mainFile;
                    await _context.SaveChangesAsync();

                    transaction.Commit();
                }
                catch (Exception ex)
                {
                    transaction.Rollback();

                    await base.HandleDbError(ex);
                    return null;
                }
            }

            #endregion

            return userSolution;
        }


        /// <summary>
        /// returns the test results (normal and custom tests) for the given release (exercise) and plang
        /// </summary>
        /// <param name="releaseCode">the exercise release</param>
        /// <param name="pLangId">the plang of the sultion</param>
        [HttpGet("testresults/{releaseCode}/{pLangId}")]
        public async Task GetAllTestResults(string releaseCode, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            await _GetAllTestResults(releaseCode, pLangId, userId);
        }

        /// <summary>
        /// no checks are done for permissions!!
        /// </summary>
        internal async Task _GetAllTestResults(string releaseCode, int pLangId, int targetUserId)
        {
            //ensure plang exists
            PLang pLang = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == pLangId);

            if (pLang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "plang not found")));
                return;
            }

            ExerciseReleaseWithUserAsParticipation releaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .FirstOrDefaultAsync(p =>
                    p.UserId == targetUserId && p.ExerciseRelease.GeneratedCode == releaseCode);

            if (releaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }

            Solution solution = await _context.Solutions
                // .Include(p => p.TestResults)
                .Include(p => p.CustomTestResults)
                .FirstOrDefaultAsync(
                    p =>
                        p.UserId == targetUserId
                        && p.ExerciseReleaseId == releaseWithUserAsParticipation.ExerciseRelease.Id
                        && p.PLangId == pLangId
                );

            var normalTestResults = await _context.TestWithSingleSolutionAsTestResult
                    .Where(p => p.Test.IsSubmitTest == false &&
                                p.SolutionUserId == solution.UserId &&
                                p.SolutionExerciseReleaseId == solution.ExerciseReleaseId &&
                                p.SolutionPLangId == solution.PLangId
                    )
                    .ToListAsync()
                ;

            if (solution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }


            var storedTestResults = new StoredTestResultTupleFromBackend
            {
                TestResults = normalTestResults.Select(p => new StoredTestResultFromBackend()
                {
                    TestId = p.TestId,
                    Passed = p.Passed,
                    ProgramExitCode = p.ProgramExitCode,
                    TestResultCode = p.TestResultCode,
                    Protocol =
                        p.Protocol.Split(new string[] {Constants.TestProtocolSplitString},
                            StringSplitOptions.RemoveEmptyEntries).ToList(),
                    HasCompiled = p.HasCompiled,

                    RunnerVersion = p.RunnerVersion,
                    TestServerVersion = p.TestServerVersion,
                    TimeForCompiling = p.TimeForCompiling,
                    TimeForUserProgram = p.TimeForUserProgram,
                    TimeoutInMsUsed = p.TimeoutInMsUsed,
                    CompileTimeoutInMsUsed = p.CompileTimeoutInMsUsed,
                }).ToList(),
                CustomTestResults = new List<StoredTestResultFromBackend>()
            };


            var exerciseTuple =
                    await _context.Exercises
                        .Where(p => p.Id == releaseWithUserAsParticipation.ExerciseRelease.ExerciseId)
                        .Select(p => new {p.CanUserCreateCustomTests})
                        .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "cannot check permission")));
                return;
            }

            if (exerciseTuple.CanUserCreateCustomTests)
            {
                //user is allowed to create custom tests --> include results

                storedTestResults.CustomTestResults =
                    solution.CustomTestResults.Select(p => new StoredTestResultFromBackend
                    {
                        TestId = p.CustomTestId,
                        Passed = p.Passed,
                        ProgramExitCode = p.ProgramExitCode,
                        TestResultCode = p.TestResultCode,
                        Protocol =
                            p.Protocol.Split(new string[] {Constants.TestProtocolSplitString},
                                StringSplitOptions.RemoveEmptyEntries).ToList(),
                        HasCompiled = p.HasCompiled,

                        RunnerVersion = p.RunnerVersion,
                        TestServerVersion = p.TestServerVersion,
                        TimeForCompiling = p.TimeForCompiling,
                        TimeForUserProgram = p.TimeForUserProgram,
                        TimeoutInMsUsed = p.TimeoutInMsUsed,
                        CompileTimeoutInMsUsed = p.CompileTimeoutInMsUsed,
                    }).ToList()
                    ;
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<StoredTestResultTupleFromBackend>(ResponseCode.Ok,
                        "", storedTestResults)));
        }

        /// <summary>
        /// returns the submit test results for the given release (exercise) and plang
        /// </summary>
        [HttpGet("testresults/submit/{releaseCode}/{pLangId}")]
        public async Task GetSubmitTestWithResults(string releaseCode, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            await _GetSubmitTestWithResultsAndAssessment(releaseCode, pLangId, userId);
        }

        /// <summary>
        /// no checks are done for permissions!!
        /// also returns the assessment & the submit tests
        /// </summary>
        /// <param name="releaseCode"></param>
        /// <param name="pLangId"></param>
        /// <param name="targetUserId"></param>
        internal async Task _GetSubmitTestWithResultsAndAssessment(string releaseCode, int pLangId, int targetUserId)
        {
            //ensure plang exists
            PLang pLang = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == pLangId);

            if (pLang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "plang not found")));
                return;
            }

            ExerciseRelease release = await _context
                .ExerciseReleases
                .FirstOrDefaultAsync(p => p.GeneratedCode == releaseCode);

            if (release == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            if (release.IsReleased)
            {
                //release must be unreleased to have assessments
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ReleaseStillActive, "release is still active")));
                return;
            }


            var solution = await _context.Solutions
                    .Include(p => p.Assessment)
                    .FirstOrDefaultAsync(
                        p => p.PLangId == pLangId
                             && p.UserId == targetUserId
                             && p.ExerciseReleaseId == release.Id)
                ;


            if (release.HasAutomaticAssessmentFinished == false)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "assessment still running")));
                return;
            }

            if (solution?.Assessment == null)
            {
                //this can happen if the solution assessment has started but has not yet finished...
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution or assessment not found")));
                return;
            }

            //check if assessment already run
            //var submitTestResults = _context.TestResults
            //    .Include(p => p.Test)
            //    .Include(p => p.Test.TestSettings)
            //    .Include(p => p.Test.TestAssets)
            //    .Where(p => p.SolutionVersionId == solution.Assessment.SolutionVersion.Id
            //                && p.Test.IsSubmitTest == true)
            //    .ToList();


            //var submitTestResultsForFrontend = submitTestResults.Select(p => new AssessedSubmitTestFromBackend()
            //{
            //    Id = p.Test.Id,
            //    Files = p.Test.TestAssets.Select(f => new TestAssetFromBackend()
            //    {
            //        Id = f.Id,
            //        DisplayName = f.DisplayName,
            //        MimeType = f.MimeType,
            //        Size = f.Content.Length
            //    }).ToList(),
            //    Weight = p.Test.Weight,
            //    IsSubmitTest = p.Test.IsSubmitTest,
            //    Content = p.Test.Content,
            //    DisplayIndex = p.Test.DisplayIndex,
            //    DisplayName = p.Test.DisplayName,
            //    TestTypeId = p.Test.TestTypeId,
            //    TestSettings = new TestSettingsFullBase()
            //    {
            //        Id = p.Test.TestSettings.Id,
            //        MaxDiskSpaceInKb = p.Test.TestSettings.MaxDiskSpaceInKb,
            //        MemoryLimitInKb = p.Test.TestSettings.MemoryLimitInKb,
            //        TimeoutInMs = p.Test.TestSettings.TimeoutInMs,
            //        CompileTimeoutInMs = p.Test.TestSettings.CompileTimeoutInMs
            //    },
            //    Result = new StoredTestResultFromBackend()
            //    {
            //        Passed = p.Passed,
            //        TestId = p.TestId,
            //        TestResultCode = p.TestResultCode,
            //        ProgramExitCode = p.ProgramExitCode,
            //        HasCompiled = p.HasCompiled,
            //        Protocol =
            //            p.Protocol.Split(new string[] { Constants.TestProtocolSplitString },
            //                StringSplitOptions.RemoveEmptyEntries).ToList()
            //    }
            //}).ToList();


            //if we have no submit tests results (e.g. not assessed or error) still show submit tests (with empty result)

            //from https://stackoverflow.com/questions/3792888/how-to-implement-left-join-in-join-extension-method
            //var submitTestsWithResults = _context.Tests
            //     .Where(p => p.IsSubmitTest == true &&
            //      release.ExerciseId == p.ExerciseId
            //     )
            //     .Include(p => p.TestSettings)
            //     .Include(p => p.TestAssets)
            //     .GroupJoin(_context.TestResults,
            //         test => test.Id,
            //         testResult => testResult.TestId && testResult.SolutionVersion.Solution.,
            //         (test, result) => new {
            //             Test = test,
            //             Result = result.DefaultIfEmpty()
            //         })
            //     .SelectMany(z => z.Result.Select(t => new { Test = z.Test, Result = t }))
            //     .ToList()
            //     ;


            //---BAD STYLE & code duplicate !!!!! but working ...
            //e.g. reuse solution from above (just include solution versions)

            var sumbmitTests = _context.Tests
                    .Where(p => p.IsSubmitTest == true &&
                                release.ExerciseId == p.ExerciseId
                    )
                    .Include(p => p.TestSettings)
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                ;


            var userSolution = await _context.Solutions
                    .FirstOrDefaultAsync(p => p.UserId == targetUserId
                                              && p.PLangId == pLangId
                                              && p.ExerciseReleaseId == release.Id
                    )
                ;

            if (userSolution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }


            var testResults = await _context.TestWithSingleSolutionAsTestResult
                    .Where(p => p.Solution == userSolution)
                    .ToListAsync()
                ;


            var submitTestsWithResults = await sumbmitTests.Select(p => new
                    {
                        Test = p,
                        Result = testResults.FirstOrDefault(r => r.TestId == p.Id)
                    })
                    .ToListAsync()
                ;


            var submitTestResultsForFrontend = submitTestsWithResults
                .Select(p => new AssessedSubmitTestFromBackend()
                {
                    Id = p.Test.Id,
                    Files = p.Test.AssetReferences.Select(fileRef => new FilePreviewFromBackend()
                    {
                        //user cannot change test assets/or view them so not include data
                        Id = fileRef.FileReferenceTestAsset.Id,
                        OriginalName = fileRef.FileReferenceTestAsset.OriginalName,
                        MimeType = fileRef.FileReferenceTestAsset.MimeType,
                        SizeInBytes = fileRef.FileReferenceTestAsset.SizeInBytes,
                    }).ToList(),
                    Weight = p.Test.Weight,
                    IsSubmitTest = p.Test.IsSubmitTest,
                    Content = p.Test.Content,
                    DisplayIndex = p.Test.DisplayIndex,
                    DisplayName = p.Test.DisplayName,
                    TestTypeId = p.Test.TestTypeId,
                    TestSettings = new TestSettingsFullBase()
                    {
                        Id = p.Test.TestSettings.Id,
                        MaxDiskSpaceInKb = p.Test.TestSettings.MaxDiskSpaceInKb,
                        MemoryLimitInKb = p.Test.TestSettings.MemoryLimitInKb,
                        TimeoutInMs = p.Test.TestSettings.TimeoutInMs,
                        CompileTimeoutInMs = p.Test.TestSettings.CompileTimeoutInMs,
                        CompilerOptions = p.Test.TestSettings.CompilerOptions
                    },
                    Result = p.Result == null
                        ? null
                        : new StoredTestResultFromBackend()
                        {
                            Passed = p.Result.Passed,
                            TestId = p.Result.TestId,
                            TestResultCode = p.Result.TestResultCode,
                            ProgramExitCode = p.Result.ProgramExitCode,
                            HasCompiled = p.Result.HasCompiled,
                            Protocol =
                                p.Result.Protocol.Split(new string[] {Constants.TestProtocolSplitString},
                                    StringSplitOptions.RemoveEmptyEntries).ToList(),

                            RunnerVersion = p.Result.RunnerVersion,
                            TestServerVersion = p.Result.TestServerVersion,
                            TimeForCompiling = p.Result.TimeForCompiling,
                            TimeForUserProgram = p.Result.TimeForUserProgram,
                            TimeoutInMsUsed = p.Result.TimeoutInMsUsed,
                            CompileTimeoutInMsUsed = p.Result.CompileTimeoutInMsUsed,
                        }
                }).ToList();

            var assessmentResult = new AssessmentResultForUser()
            {
                FeedbackForStudent = solution.Assessment?.FeedbackForStudent,
                ManualPoints = solution.Assessment?.ManualPoints,
                MaxNormalTestPoints = solution.Assessment?.MaxNormalTestPoints,
                MaxSubmitTestPoints = solution.Assessment?.MaxSubmitTestPoints,
                NormalTestPoints = solution.Assessment?.NormalTestPoints,
                SubmitTestPoints = solution.Assessment?.SubmitTestPoints,
                UserId = solution.UserId,
                ReleaseId = userSolution.ExerciseReleaseId,
                PLangId = userSolution.PLangId
            };

            var entity = new AssessmentForUserWithTestResults()
            {
                SubmitTestResults = submitTestResultsForFrontend,
                AssessmentResult = solution.Assessment != null ? assessmentResult : null
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<AssessmentForUserWithTestResults>(ResponseCode.Ok,
                        "submit test results queried", entity)));
        }

//        /// <summary>
//        /// returns a new solution version
//        /// </summary>
//        /// <param name="solutionForBackend">the solution tuple from the frontend (new files)</param>
//        /// <param name="oldCodeTemplateForPLang">the code template for the plang (used to check readonly files)</param>
//        /// <param name="oldRelease">the release (to check if the user was allowed to add files)</param>
//        /// <param name="mainClassFileTuple">the main class file (this should be set in this method to later save the main file)</param>
//        /// <param name="fileAssociation">the list of frontend/backend associated files to track the file ids,
//        ///  to know hich frontend file belongs to which backend file</param>
//        /// <returns>null (no main file found) or the new solution</returns>
//        private SolutionVersion SaveNewSolutionVersion(SolutionDoExerciseFullBase solutionForBackend,
//            CodeTemplate oldCodeTemplateForPLang, ExerciseRelease oldRelease,
//            ref Tuple<SolutionVersion, SolutionFile> mainClassFileTuple,
//            out List<Tuple<SolutionFileDoExerciseFullBase, SolutionFile>> fileAssociation)
//        {
//            fileAssociation = new List<Tuple<SolutionFileDoExerciseFullBase, SolutionFile>>();
//
//            #region save new version
//
//            //ignore main file id form frontend and find the main file id through the connected code templates 
//            //(connected to the solition files) because we created the solution files from the templates
//
//            SolutionFileDoExerciseFullBase mainClassFile = null;
//
//            foreach (var solutionFile in solutionForBackend.SolutionFiles)
//            {
//                //files created by the user does not have a tempalte ... so check this first
//                if (solutionFile.TemplateFileId != null)
//                {
//                    //get the template from which we created the solution file 
//                    var oldTemplateFile =
//                        oldCodeTemplateForPLang.TemplateFiles.FirstOrDefault(p => solutionFile.TemplateFileId == p.Id);
//
//                    if (oldTemplateFile != null)
//                    {
//                        //we found the old template... is this the main file of the code template for the plang?
//
//                        if (oldCodeTemplateForPLang.MainFile == oldTemplateFile)
//                        {
//                            //we found the main file ... use this solution file as the main one because it was created 
//                            //from the main code template file
//                            mainClassFile = solutionFile;
//                            break;
//                        }
//                    }
//                }
//            }
//
//
//            //this is unsave but maybe one day the user can set the main file by him/herself...
//            //SolutionFileDoExerciseFullBase mainClassFile =
//            //    solutionVersionForBackend.SolutionFiles.FirstOrDefault(
//            //        p => p.Id == solutionVersionForBackend.MainFileId);
//
//            if (mainClassFile == null)
//            {
//                return null;
//            }
//
//
//            //first save
//
//            var newSolutionVersion = new SolutionVersion()
//            {
//                SolutionFiles = new List<SolutionFile>(),
//                CreatedAt = DateTimeHelper.GetUtcNow()
//                //because CreatedAt is the same name as the shadow propertsy CreatedAt this is set by the entity framework...
//                //but we need this in order to sort the solution versions (to delete the old ones)
//                //so the real create at value will be different (a tiny bit later but this does not matter for the sorting)
//            };
//
//
//            //check every file if the user is allowed to add
//            foreach (SolutionFileDoExerciseFullBase solutionFile in solutionForBackend.SolutionFiles)
//            {
//                TemplateFile oldTemplate =
//                    oldCodeTemplateForPLang.TemplateFiles.FirstOrDefault(p => p.Id == solutionFile.TemplateFileId);
//
//
//                if (oldTemplate == null)
//                {
//                    //this is a file the user has created (else a template is present)
//                    //user created new file... check if allowed...
//                    if (oldRelease.Exercise.CanUserCreateFiles)
//                    {
//                        var part = new SolutionFile()
//                        {
//                            Content = solutionFile.Content,
//                            FileNameWithExtension = solutionFile.FileNameWithExtension,
//                            TemplateFileId = null,
//                            IsDisplayed = solutionFile.UiIsDisplayed
//                        };
//
//                        newSolutionVersion.SolutionFiles.Add(part);
//                        fileAssociation.Add(
//                            new Tuple<SolutionFileDoExerciseFullBase, SolutionFile>(solutionFile, part));
//
//                        //this cannot be the main class file (because the main file must be set when creating the exercise...)
//                        //MAYBE possible in further versions??
//                    }
//                    else
//                    {
//                        //ignore file because user was not allowed to...
//                    }
//                }
//                else
//                {
//                    //create a new file (template exists)
//
//                    bool isMainFile = oldCodeTemplateForPLang.MainFile.Id == oldTemplate.Id;
//
//                    SolutionFile file = null;
//
//                    if (oldTemplate.IsEditableByUser)
//                    {
//                        file = new SolutionFile()
//                        {
//                            Content = solutionFile.Content,
//                            FileNameWithExtension = solutionFile.FileNameWithExtension,
//                            TemplateFile = oldTemplate,
//                            IsDisplayed = solutionFile.UiIsDisplayed
//                        };
//                        fileAssociation.Add(
//                            new Tuple<SolutionFileDoExerciseFullBase, SolutionFile>(solutionFile, file));
//                    }
//                    else
//                    {
//                        //better use the file from the template here...
//                        file = new SolutionFile()
//                        {
//                            Content = oldTemplate.Content,
//                            FileNameWithExtension = oldTemplate.FileNameWithExtension,
//                            TemplateFile = oldTemplate,
//                            IsDisplayed = solutionFile.UiIsDisplayed
//                        };
//                        fileAssociation.Add(
//                            new Tuple<SolutionFileDoExerciseFullBase, SolutionFile>(solutionFile, file));
//                    }
//
//                    //now add file
//                    newSolutionVersion.SolutionFiles.Add(file);
//
//                    if (isMainFile)
//                    {
//                        // oldSolution.Solution.MainClassFile = part;
//                        //cannot save because ef circular dependency --> save after all files are saved
//                        mainClassFileTuple = new Tuple<SolutionVersion, SolutionFile>(newSolutionVersion, file);
//                    }
//                }
//            }
//
//            return newSolutionVersion;
//
//            #endregion
//        }

        /// <summary>
        /// overwrites the given solution (this also deletes the old test results (because with new files they are invalid))
        /// custom & normal tests needs to be included else we cannot clear them
        /// </summary>
        /// <param name="solutionForBackend">the solution tuple from the frontend (new files)</param>
        /// <param name="oldSolution">the old solution version to overwrite</param>
        /// <param name="oldCodeTemplateForPLang">the code template for the plang (used to check readonly files)</param>
        /// <param name="oldRelease">the release (to check if the user was allowed to add files)</param>
        /// <param name="shouldClearAllTestResults">true: clear all test results, false: not (the results must be loaded!!)</param>
        /// <param name="mainClassFileTuple">the main class file (this should be set in this method to later save the main file)</param>
        /// <param name="fileAssociation">to get the new files ids in case we created some</param>
        /// <returns>null: all ok, string: the invalid file name</returns>
        private string OverwriteSolution(SolutionDoExerciseFullBase solutionForBackend,
            Solution oldSolution,
            CodeTemplate oldCodeTemplateForPLang, ExerciseRelease oldRelease,
            bool shouldClearAllTestResults,
            ref Tuple<Solution, SolutionFile> mainClassFileTuple,
            out List<Tuple<SolutionFileDoExerciseFullBase, SolutionFile>> fileAssociation
        )
        {
            var addedSolutionFiles = new List<SolutionFile>();
            fileAssociation = new List<Tuple<SolutionFileDoExerciseFullBase, SolutionFile>>();

            /*
             * disallow adding new template files OR add new template files if we have releases...
             * else on save files we cannot decide if an old file should be removed or added
             * (foreach old file check if frontend version exists... if no? create [new] template file for user or delete?)
             * 
             * new solutions will be initiated with all templates so no problem there
             */

            #region delete removed files

            for (var i = 0; i < oldSolution.SolutionFiles.Count; i++)
            {
                var oldFile = oldSolution.SolutionFiles[i];
                var frontendFile = solutionForBackend.SolutionFiles.FirstOrDefault(p => p.Id == oldFile.Id);

                if (frontendFile != null)
                {
                    //the old file is still present in the new frontend version
                    continue;
                }

                if (oldFile.TemplateFileId.HasValue)
                {
                    //file was created from template... cannot be deleted by user
                    return oldFile.FileNameWithExtension + " cannot be deleted because it was created from a template";
                }

                //old file should be removed
                oldSolution.SolutionFiles.Remove(oldFile);
                i--;
            }

            #endregion

            #region old version exists

            //main class file cannot change...
            //MAYBE possible in further versions??

            foreach (SolutionFileDoExerciseFullBase solutionFile in solutionForBackend.SolutionFiles)
            {
                SolutionFile oldFile = oldSolution.SolutionFiles.FirstOrDefault(p => p.Id == solutionFile.Id);

                TemplateFile oldTemplate = null;

                if (oldFile == null)
                {
                    //no solution file exists yet (maybe the user created new one for old solution)
                    //or the tutor added a new file in the templates

                    if (solutionFile.TemplateFileId.HasValue)
                    {
                        oldTemplate =
                            oldCodeTemplateForPLang.TemplateFiles.FirstOrDefault(
                                p => p.Id == solutionFile.TemplateFileId.Value);
                    }

                    SolutionFile file = null;

                    if (oldTemplate == null) //this is a file the user has created (else a template is present)
                    {
                        //user created new file... check if allowed...
                        if (oldRelease.Exercise.CanUserCreateFiles)
                        {
                            if (Files.ValidFileName(solutionFile.FileNameWithExtension.Trim()) == false)
                            {
                                //not a valid file name...
                                return solutionFile.FileNameWithExtension.Trim();
                            }

                            file = new SolutionFile()
                            {
                                Content = solutionFile.Content,
                                FileNameWithExtension = solutionFile.FileNameWithExtension.Trim(),
                                IsDisplayed = solutionFile.UiIsDisplayed
                            };

                            oldSolution.SolutionFiles.Add(file);
                            addedSolutionFiles.Add(file);
                            fileAssociation.Add(
                                new Tuple<SolutionFileDoExerciseFullBase, SolutionFile>(solutionFile, file));
                        }
                        else
                        {
                            //ignore file because user was not allowed to create...
                        }
                    }
                    else
                    {
                        //create a new file (template exists)

                        if (oldTemplate.IsEditableByUser) //use user content
                        {
                            file = new SolutionFile()
                            {
                                Content = solutionFile.Content,
                                FileNameWithExtension = solutionFile.FileNameWithExtension,
                                TemplateFile = oldTemplate,
                                IsDisplayed = solutionFile.UiIsDisplayed
                            };
                            oldSolution.SolutionFiles.Add(file);
                            addedSolutionFiles.Add(file);
                            fileAssociation.Add(
                                new Tuple<SolutionFileDoExerciseFullBase, SolutionFile>(solutionFile, file));
                        }
                        else
                        {
                            //use template content (because user should not modify content)
                            //better use the file from the template here...
                            file = new SolutionFile()
                            {
                                Content = oldTemplate.Content,
                                FileNameWithExtension = oldTemplate.FileNameWithExtension,
                                TemplateFile = oldTemplate,
                                IsDisplayed = solutionFile.UiIsDisplayed
                            };
                            oldSolution.SolutionFiles.Add(file);
                            addedSolutionFiles.Add(file);
                            fileAssociation.Add(
                                new Tuple<SolutionFileDoExerciseFullBase, SolutionFile>(solutionFile, file));
                        }
                    }

                    if (solutionForBackend.MainFileId == solutionFile.Id && file != null)
                    {
                        mainClassFileTuple = new Tuple<Solution, SolutionFile>(oldSolution, file);
                    }
                }
                else
                {
                    oldTemplate = oldFile.TemplateFile;

                    //overwrite old file
                    if (oldTemplate == null)
                    {
                        //this is a file the user has created (else a template is present) so he can edit this 
                        oldFile.Content = solutionFile.Content;
                        oldFile.FileNameWithExtension = solutionFile.FileNameWithExtension;
                        oldFile.IsDisplayed = solutionFile.UiIsDisplayed;
                    }
                    else
                    {
                        if (oldTemplate.IsEditableByUser) //only modify if allowed
                        {
                            oldFile.Content = solutionFile.Content;
                            oldFile.FileNameWithExtension = solutionFile.FileNameWithExtension;
                            oldFile.IsDisplayed = solutionFile.UiIsDisplayed;
                        }
                        else
                        {
                            //better reset to the template file
                            //TODO or should we?? see issue #56 frontend
                            oldFile.Content = oldTemplate.Content;
                            oldFile.FileNameWithExtension = oldTemplate.FileNameWithExtension;
                            oldFile.IsDisplayed = solutionFile.UiIsDisplayed;
                        }
                    }

                    fileAssociation.Add(
                        new Tuple<SolutionFileDoExerciseFullBase, SolutionFile>(solutionFile, oldFile));

                    //Main class does not change ... but user files could be invalid
                    if (oldFile.Id == oldSolution.MainFileId)
                    {
                        mainClassFileTuple = new Tuple<Solution, SolutionFile>(oldSolution, oldFile);
                    }
                }
            }

            #endregion

            //only works when the test results are included from the db
            if (shouldClearAllTestResults)
            {
                oldSolution.CustomTestResults.Clear();
                oldSolution.TestResults.Clear();
            }

            oldSolution.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            return null;
        }

        /// <summary>
        /// saves the user solution for the given exercise and release and plang
        /// 
        /// <para></para>
        /// IF this method returns an InvalidRequest and a string, then a file name was invalid... the string contains the invalid file name
        /// </summary>
        /// <param name="generatedCode">the code for the exercise release</param>
        /// <param name="shouldClearAllTestResults">let the frontend decide if the code has changed... (e.g. editor has unsaved changes)</param>
        /// <param name="solutionForBackend">the exercise and release and plang</param>
        [HttpPost("save/{generatedCode}/{shouldClearAllTestResults}")]
        public async Task SaveCodeFiles(string generatedCode, bool shouldClearAllTestResults,
            [FromBody] SolutionDoExerciseFullBase solutionForBackend)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (solutionForBackend == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no files found")));
                return;
            }


            //check if p lang is valid

            if ((await _context.PLangs.FirstOrDefaultAsync(p => p.Id == solutionForBackend.PLangId)) == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang found")));
                return;
            }

            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation =
                    await _context.ExerciseReleaseWithUserAsParticipations
                        .Where(p => p.ExerciseRelease.GeneratedCode == generatedCode && p.UserId == userId)
                        .Include(p => p.Solutions)
                        .ThenInclude(p => p.SolutionFiles)
                        .Include(p => p.Solutions)
                        .ThenInclude(p => p.CustomTestResults)
                        .Include(p => p.Solutions)
                        .ThenInclude(p => p.TestResults)
                        .Include(p => p.ExerciseRelease)
                        .Include(p => p.ExerciseRelease.Exercise)
                        .ThenInclude(exercise => exercise.CodeTemplates)
                        .ThenInclude(p => p.TemplateFiles)
                        .FirstOrDefaultAsync()
                ;

            ExerciseRelease release = oldExerciseReleaseWithUserAsParticipation?.ExerciseRelease;


            if (oldExerciseReleaseWithUserAsParticipation?.ExerciseRelease == null)
            {
                //this should not happen because when accessing the exercise (getting exercise) then the participation is ensured...

                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            if (release.Exercise == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                        "exercise was not found")));
                return;
            }

            if (release.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            //check if available working time is up
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }

            if (oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }


            oldExerciseReleaseWithUserAsParticipation.LastEditedPLangId =
                solutionForBackend.PLangId;


            //use template to get readonly files and check if youser can modify files
            CodeTemplate oldCodeTemplateForPLang = release.Exercise.CodeTemplates
                .FirstOrDefault(p => p.PLangId == solutionForBackend.PLangId);


            if (oldCodeTemplateForPLang == null)
            {
                //no template exists...
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no template found")));
                return;
            }


            //search for solution for the given p lang
            Solution oldSolution =
                oldExerciseReleaseWithUserAsParticipation.Solutions
                    .FirstOrDefault(
                        p => p.PLangId == solutionForBackend.PLangId);


            if (oldSolution == null)
            {
                //this should not happen because when accessing the exercise (getting exercise) then the participation (and solution) is ensured...
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }


            List<Tuple<SolutionFileDoExerciseFullBase, SolutionFile>> fileAssociation = null;
            Tuple<Solution, SolutionFile> mainClassFileTuple = null;

            #region handle solution update

            string invalidFileName = OverwriteSolution(solutionForBackend,
                oldSolution, oldCodeTemplateForPLang,
                release, shouldClearAllTestResults, ref mainClassFileTuple, out fileAssociation);

            if (invalidFileName != null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                    "invalid filename: " + invalidFileName)));
                return;
            }

            oldSolution.LastEditingIpAddress = NetworkHelper.GetIpAddress(HttpContext);

            oldExerciseReleaseWithUserAsParticipation.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            #endregion


            //check if main file really got set
            if (mainClassFileTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no main file found")));
                return;
            }

            using (var transaction = _context.Database.BeginTransaction())
            {
                try
                {
                    //use a transaction to ensure the main file is already 

                    //save the solution changes
                    await _context.SaveChangesAsync();

                    //always save in case user has an invalid solution
                    mainClassFileTuple.Item1.MainFile = mainClassFileTuple.Item2;
                    await _context.SaveChangesAsync();

                    transaction.Commit();
                }
                catch (Exception ex)
                {
                    transaction.Rollback();

                    await base.HandleDbError(ex);
                    return;
                }
            }


            var solutionVersionForFrontend = new SolutionDoExerciseFullBase()
            {
                UserId = oldSolution.UserId,
                ReleaseId = oldSolution.ExerciseReleaseId,
                PLangId = oldSolution.PLangId,
                LastUpdatedAt = oldSolution.LastUpdatedAt,
                // ReSharper disable once PossibleInvalidOperationException //this should be set or checked in OverwriteSolutionVersion or SaveNewSolutionVersion
                MainFileId = oldSolution.MainFileId ?? 0,
                SolutionFiles =
                    oldSolution.SolutionFiles.Select(
                        p => new SolutionFileDoExerciseFullBase() //solutionVersionForBackend
                        {
                            Id = p.Id,
                            Content = p.Content,
                            UiIsDisplayed = p.IsDisplayed,
                            //when there is no template file the user created the file and is allowed to edit it
                            IsEditableByUser = (p.TemplateFile == null || p.TemplateFile.IsEditableByUser),
                            FileNameWithExtension = p.FileNameWithExtension,
                            TemplateFileId = p.TemplateFileId,
                            DisplayIndex = p.DisplayIndex,
                            //CodeTemplateId = oldCodeTemplateForPLang.Id //TODO delete if not used
                        }).ToList()
            };


            var solutionVersionTupleForFrontend = new SolutionTupleFromBackend
            {
                Solution = solutionVersionForFrontend,
                FileIdAssociationMap = fileAssociation.Select(p => new FileIdAssociation()
                {
                    FrontendFileId = p.Item1.Id,
                    BackendFileId = p.Item2.Id
                }).ToList()
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<SolutionTupleFromBackend>(ResponseCode.Ok,
                        "saved solution version",
                        solutionVersionTupleForFrontend))
                );
        }


        [HttpPost("reset/{generatedCode}/{pLangId}")]
        public async Task ResetCode(string generatedCode, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            //check if p lang is valid

            PLang pLang = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == pLangId);

            if (pLang == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang found")));
                return;
            }

            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation =
                    await _context.ExerciseReleaseWithUserAsParticipations
                        .Where(p => p.ExerciseRelease.GeneratedCode == generatedCode && p.UserId == userId)
                        .Include(p => p.Solutions)
                        .ThenInclude(p => p.SolutionFiles)
                        .Include(p => p.Solutions)
                        .ThenInclude(p => p.CustomTestResults)
                        .Include(p => p.Solutions)
                        .ThenInclude(p => p.TestResults)
                        .Include(p => p.ExerciseRelease)
                        .Include(p => p.ExerciseRelease.Exercise)
                        .ThenInclude(exercise => exercise.CodeTemplates)
                        .ThenInclude(p => p.TemplateFiles)
                        .FirstOrDefaultAsync()
                ;

            ExerciseRelease release = oldExerciseReleaseWithUserAsParticipation?.ExerciseRelease;

            if (oldExerciseReleaseWithUserAsParticipation?.ExerciseRelease == null)
            {
                //this should not happen because when accessing the exercise (getting exercise) then the participation is ensured...

                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            if (release.Exercise == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (release.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            //check if available working time is up
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }

            if (oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }


            oldExerciseReleaseWithUserAsParticipation.LastEditedPLangId = pLangId;


            //use template to get readonly files and check if youser can modify files
            CodeTemplate oldCodeTemplateForPLang = release.Exercise.CodeTemplates
                .FirstOrDefault(p => p.PLangId == pLangId);


            if (oldCodeTemplateForPLang == null)
            {
                //no template exists...
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no template found")));
                return;
            }


            //search for solution for the given p lang
            Solution oldSolution =
                oldExerciseReleaseWithUserAsParticipation.Solutions
                    .FirstOrDefault(
                        p => p.PLangId == pLangId);

            if (oldSolution == null)
            {
                //this should not happen because when accessing the exercise (getting exercise) then the participation (and solution) is ensured...
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }


            //all checks are done... create a new initial solution


            Solution resetSolution = await _CreateInitialSolution(oldSolution, release, pLang, userId,
                oldExerciseReleaseWithUserAsParticipation);

            if (resetSolution == null) return;


            var solutionVersionForFrontend = new SolutionDoExerciseFullBase()
            {
                UserId = resetSolution.UserId,
                ReleaseId = resetSolution.ExerciseReleaseId,
                PLangId = resetSolution.PLangId,
                LastUpdatedAt = resetSolution.LastUpdatedAt,
                // ReSharper disable once PossibleInvalidOperationException //this should be set or checked in OverwriteSolutionVersion or SaveNewSolutionVersion
                MainFileId = resetSolution.MainFileId ?? 0,
                SolutionFiles =
                    resetSolution.SolutionFiles.Select(
                        p => new SolutionFileDoExerciseFullBase()
                        {
                            Id = p.Id,
                            Content = p.Content,
                            UiIsDisplayed = p.IsDisplayed,
                            //when there is no template file the user created the file and is allowed to edit it
                            IsEditableByUser = (p.TemplateFile == null || p.TemplateFile.IsEditableByUser),
                            FileNameWithExtension = p.FileNameWithExtension,
                            TemplateFileId = p.TemplateFileId,
                            DisplayIndex = p.DisplayIndex,
                        }).ToList()
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<SolutionDoExerciseFullBase>(ResponseCode.Ok,
                        "reset code",
                        solutionVersionForFrontend))
                );
        }


        /// <summary>
        /// sets the lock solutions flag on the participation
        /// </summary>
        /// <param name="generatedCode"></param>
        /// <param name="shouldLock"></param>
        /// <returns></returns>
        [HttpPost("lock/solutions/{generatedCode}/{shouldLock}")]
        public async Task SetLockSolutions(string generatedCode, bool shouldLock)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation =
                    await _context.ExerciseReleaseWithUserAsParticipations
                        .Include(p => p.ExerciseRelease)
                        .ThenInclude(p => p.Exercise)
                        .FirstOrDefaultAsync(p =>
                            p.ExerciseRelease.GeneratedCode == generatedCode && p.UserId == userId)
                ;

            if (oldExerciseReleaseWithUserAsParticipation == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }


            if (oldExerciseReleaseWithUserAsParticipation.ExerciseRelease == null)
            {
                //this should not happen because when accessing the exercise (getting exercise) then the participation is ensured...

                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            if (oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                        "exercise was not found")));
                return;
            }

            if (oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            //check if available working time is up
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }

            if (oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.Exercise.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }


            oldExerciseReleaseWithUserAsParticipation.LockSolutionsFlag = shouldLock;

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<bool>(ResponseCode.Ok,
                        "",
                        oldExerciseReleaseWithUserAsParticipation.LockSolutionsFlag))
                );
        }

        /// <summary>
        /// creates a new custom test for a exercise (and release)
        /// </summary>
        /// <param name="generatedCode">the code from the exercise release</param>
        /// <param name="customTestForBackend">the new values</param>
        [HttpPost("create/customtest/{generatedCode}")]
        public async Task CreateCustomTest(string generatedCode,
            [FromBody] CustomTestForBackend customTestForBackend)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (customTestForBackend == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no test found")));
                return;
            }


            var oldExerciseReleaseWithUserAsSolutionWithExerciseId = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Where(p => p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode)
                .Include(p => p.ExerciseRelease)
                .Select(p => new
                {
                    OldExerciseReleaseConnection = p,
                    ExerciseId = p.ExerciseRelease.ExerciseId,
                    CustomTestsCount = p.CustomTests.Count
                })
                .FirstOrDefaultAsync();


            var exerciseTuple =
                    await _context.Exercises
                        .Where(p => p.Id == oldExerciseReleaseWithUserAsSolutionWithExerciseId.ExerciseId)
                        .Select(p => new
                            {p.CanUserCreateCustomTests, p.IsPermanentlyLocked, p.IsOnlyVisibleToOwner, p.UserId})
                        .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (exerciseTuple.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (exerciseTuple.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exerciseTuple.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            if (exerciseTuple.CanUserCreateCustomTests == false)
            {
                //user is not allowed to create custom tests

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.InvalidRequest, "not allowed to create custom test")));
                return;
            }


            TestType testType = await _context.TestTypes
                .FirstOrDefaultAsync(p => p.Id == customTestForBackend.TestTypeId);

            if (testType == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no test type found")));
                return;
            }

            if (oldExerciseReleaseWithUserAsSolutionWithExerciseId == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsSolution =
                oldExerciseReleaseWithUserAsSolutionWithExerciseId.OldExerciseReleaseConnection;

            int oldExerciseId = oldExerciseReleaseWithUserAsSolutionWithExerciseId.ExerciseId;


            if (oldExerciseReleaseWithUserAsSolution == null)
            {
                //this should not happen because getting the exercise ensures the participation (and solution and one solution version)
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }


            //check if available working time is up
            //do not allow adding/changing custom tests after release finished (currently)
            //because we would invalidate stored custom test results
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsSolution,
                oldExerciseReleaseWithUserAsSolution.ExerciseRelease, this))
            {
                return;
            }


            //get DefaultCustomTestCaseSettings
            DefaultCustomTestSettings defaultCustomTestCaseSettings =
                await _context.DefaultCustomTestSettings.FirstOrDefaultAsync(p => p.ExerciseId == oldExerciseId);


            if (defaultCustomTestCaseSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "default test settings not found")));
                return;
            }

            var systemSetting = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSetting == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no system settings found")));
                return;
            }

            //systemSetting.MaxCustomTestsPerParticipation < 0 is unlimited
            if (systemSetting.MaxCustomTestsPerParticipation == 0 ||
                (systemSetting.MaxCustomTestsPerParticipation > 0 &&
                 oldExerciseReleaseWithUserAsSolutionWithExerciseId.CustomTestsCount + 1 >
                 systemSetting.MaxCustomTestsPerParticipation))
            {
                //not allowed
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "too much custom tests, max allowed is " + systemSetting.MaxCustomTestsPerParticipation)));
                return;
            }

            var newCustomTestCase = new CustomTest()
            {
                Content = customTestForBackend.Content,
                DisplayName = customTestForBackend.DisplayName.Trim(),
                DisplayIndex = customTestForBackend.DisplayIndex,
                TestTypeId = testType.Id,
                UserId = userId,
                ExerciseReleaseId = oldExerciseReleaseWithUserAsSolution.ExerciseReleaseId
            };

            oldExerciseReleaseWithUserAsSolution.CustomTests.Add(newCustomTestCase);
            oldExerciseReleaseWithUserAsSolution.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }


            var customTestForFrontend = new CustomTestFromBackend
            {
                Id = newCustomTestCase.Id,
                DisplayName = newCustomTestCase.DisplayName,
                Content = newCustomTestCase.Content,
                DisplayIndex = newCustomTestCase.DisplayIndex,
                TestTypeId = newCustomTestCase.TestTypeId,
                Files = new List<FilePreviewFromBackend>()
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<CustomTestFromBackend>(ResponseCode.Ok,
                        "",
                        customTestForFrontend)));
        }


        /// <summary>
        /// changes a custom test
        /// </summary>
        /// <param name="generatedCode">the code from the exercise release</param>
        /// <param name="customTestForBackend">the new values</param>
        [HttpPost("update/customtest/{generatedCode}")]
        public async Task ChangeCustomTest(string generatedCode,
            [FromBody] CustomTestForBackend customTestForBackend)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (customTestForBackend == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no test found")));
                return;
            }


            TestType testType =
                await _context.TestTypes.FirstOrDefaultAsync(p => p.Id == customTestForBackend.TestTypeId);

            if (testType == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no test type found")));
                return;
            }


            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .Include(p => p.CustomTests)
                .ThenInclude(p => p.TestResultsNew)
                .Include(p => p.CustomTests)
                .ThenInclude(p => p.AssetReferences)
                .ThenInclude(p => p.FileReferenceUserFileAsset)
                .FirstOrDefaultAsync(p =>
                    p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode);


            if (oldExerciseReleaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }

            //check if available working time is up
            //do not allow adding/changing custom tests after release finished (currently)
            //because we would invalidate stored custom test results
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }


            var exerciseTuple =
                    await _context.Exercises
                        .Where(p => p.Id == oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.ExerciseId)
                        .Select(p => new
                            {p.CanUserCreateCustomTests, p.IsPermanentlyLocked, p.IsOnlyVisibleToOwner, p.UserId})
                        .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (exerciseTuple.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (exerciseTuple.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exerciseTuple.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }


            if (exerciseTuple.CanUserCreateCustomTests == false)
            {
                //user is not allowed to create custom tests

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.InvalidRequest, "not allowed to update custom test")));
                return;
            }


            CustomTest oldCustomTest =
                oldExerciseReleaseWithUserAsParticipation.CustomTests.FirstOrDefault(
                    p => p.Id == customTestForBackend.Id);


            if (oldCustomTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom test not found")));
                return;
            }


            oldCustomTest.Content = customTestForBackend.Content;
            oldCustomTest.DisplayName = customTestForBackend.DisplayName.Trim();
            oldCustomTest.TestTypeId = testType.Id;


            //delete old (now invalid) test results...
            //should work if the test results are loaded from the db
            oldCustomTest.TestResultsNew.Clear();
            oldExerciseReleaseWithUserAsParticipation.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }


            var customTestForFrontend = new CustomTestFromBackend
            {
                Id = oldCustomTest.Id,
                DisplayName = oldCustomTest.DisplayName,
                Content = oldCustomTest.Content,
                DisplayIndex = oldCustomTest.DisplayIndex,
                TestTypeId = oldCustomTest.TestTypeId,
                Files = oldCustomTest.AssetReferences.Select(p => new FilePreviewFromBackend()
                {
                    Id = p.FileReferenceUserFileAsset.Id,
                    OriginalName = p.FileReferenceUserFileAsset.OriginalName,
                    MimeType = p.FileReferenceUserFileAsset.MimeType,
                    SizeInBytes = p.FileReferenceUserFileAsset.SizeInBytes,
                }).ToList()
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<CustomTestFromBackend>(ResponseCode.Ok,
                        "custom test changed",
                        customTestForFrontend)));
        }


        /// <summary>
        /// updates the custom test indices
        /// </summary>
        /// <param name="generatedCode">the code from the exercise release</param>
        /// <param name="tuples">the tuples with the test id and the new display index</param>
        [HttpPost("update/indices/customtest/{generatedCode}")]
        public async Task UpdateCustomTestDisplayIndices(string generatedCode,
            [FromBody] UpdateCustomTestIndicesTupleWrapper tuples)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (tuples == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "no tuples received")));
                return;
            }


            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .Include(p => p.CustomTests)
                .FirstOrDefaultAsync(p =>
                    p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode);


            if (oldExerciseReleaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }

            //check if available working time is up
            //do not allow adding/changing custom tests after release finished (currently)
            //because we would invalidate stored custom test results
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }


            var exerciseTuple =
                    await _context.Exercises
                        .Where(p => p.Id == oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.ExerciseId)
                        .Select(p => new
                            {p.CanUserCreateCustomTests, p.IsPermanentlyLocked, p.IsOnlyVisibleToOwner, p.UserId})
                        .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (exerciseTuple.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (exerciseTuple.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exerciseTuple.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            if (exerciseTuple.CanUserCreateCustomTests == false)
            {
                //user is not allowed to create custom tests

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.InvalidRequest, "not allowed to update custom test")));
                return;
            }

            foreach (var tuple in tuples.Tuples)
            {
                var oldCustomTest =
                    oldExerciseReleaseWithUserAsParticipation.CustomTests.FirstOrDefault(p => p.Id == tuple.Id);

                if (oldCustomTest == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom test not found")));
                    return;
                }

                oldCustomTest.DisplayIndex = tuple.NewDisplayIndex;
            }

            oldExerciseReleaseWithUserAsParticipation.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await Response.WriteAsync(
                Jc.Serialize(new BasicResponse(ResponseCode.Ok, "indices ensured")));
        }


        /// <summary>
        /// removes the given custom test (by id)
        /// </summary>
        /// <param name="generatedCode">the code from the exercise release</param>
        /// <param name="customTestId">the id of the custom test to remove</param>
        [HttpDelete("delete/customtest/{generatedCode}/{customTestId}")]
        public async Task DeleteCustomTest(string generatedCode, int customTestId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .Include(p => p.CustomTests)
                .FirstOrDefaultAsync(p =>
                    p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode);


            if (oldExerciseReleaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }

            //check if available working time is up
            //do not allow adding/changing custom tests after release finished (currently)
            //because we would invalidate stored custom test results
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }

            var exerciseTuple =
                    await _context.Exercises
                        .Where(p => p.Id == oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.ExerciseId)
                        .Select(p => new
                            {p.CanUserCreateCustomTests, p.IsPermanentlyLocked, p.IsOnlyVisibleToOwner, p.UserId})
                        .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (exerciseTuple.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (exerciseTuple.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exerciseTuple.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            if (exerciseTuple.CanUserCreateCustomTests == false)
            {
                //user is not allowed to create custom tests

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.InvalidRequest, "not allowed to delete custom test")));
                return;
            }

            var oldCustomTest = oldExerciseReleaseWithUserAsParticipation.CustomTests
                .FirstOrDefault(p => p.Id == customTestId);

            if (oldCustomTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom test not found")));
                return;
            }

            oldExerciseReleaseWithUserAsParticipation.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                oldExerciseReleaseWithUserAsParticipation.CustomTests.Remove(oldCustomTest);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "custom test deleted")));
        }

        //TODO maybe function to download custom test asset


        #region custom test asset handling

        /// <summary>
        /// this will upload (receive) the file and connect it to the custom test
        /// </summary>
        /// <param name="generatedCode"></param>
        /// <param name="customTestId"></param>
        /// <param name="file"></param>
        /// <returns></returns>
        [HttpPost("upload/create/custom/tests/{generatedCode}/{customTestId}")]
        public async Task UploadCustomTestAssetFile(string generatedCode, int customTestId, IFormFile file)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .Include(p => p.CustomTests)
                .FirstOrDefaultAsync(p =>
                    p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode);


            if (oldExerciseReleaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }

            var exerciseTuple = await _context.Exercises
                    .Where(p => p.Id == oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.ExerciseId)
                    .Select(p => new
                    {
                        p.CanUserCreateCustomTests,
                        p.IsPermanentlyLocked,
                        p.IsOnlyVisibleToOwner,
                        p.UserId
                    })
                    .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (exerciseTuple.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (exerciseTuple.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exerciseTuple.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            if (exerciseTuple.CanUserCreateCustomTests == false)
            {
                //user is not allowed to create custom tests

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.InvalidRequest, "not allowed to create custom test")));
                return;
            }

            //check if available working time is up
            //do not allow adding/changing custom tests after release finished (currently)
            //because we would invalidate stored custom test results
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }


            if (file == null || file.Length == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no files selected")));
                return;
            }

            var oldTest = oldExerciseReleaseWithUserAsParticipation.CustomTests
                .FirstOrDefault(p => p.Id == customTestId
                                     && p.ExerciseReleaseId ==
                                     oldExerciseReleaseWithUserAsParticipation.ExerciseReleaseId);

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.UserAssets);
            FileReferenceUserFileAsset fileRef = null;
            try
            {
                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        fileRef = new FileReferenceUserFileAsset()
                        {
                            Hash = "",
                            MimeType = file.ContentType,
                            OriginalName = file.FileName,
                        };

                        var connection = new CustomTestWithFileAsAssetReference()
                        {
                            FileReferenceUserFileAsset = fileRef,
                            CustomTest = oldTest
                        };

                        _context.CustomTestWithFileAsAssetReferences.Add(connection);
                        await _context.SaveChangesAsync();

                        Tuple<string, long> _tuple =
                            await Files.CreateUploadedFile(targetHardDrivePath, fileRef.Id, file);
                        fileRef.Hash = _tuple.Item1;
                        fileRef.SizeInBytes = _tuple.Item2;

                        await _context.SaveChangesAsync();
                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }


            var assetPreviewForFrontend = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                MimeType = fileRef.MimeType,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "",
                            assetPreviewForFrontend)));
        }

        /// <summary>
        /// adds an already uploaded file to the custom test
        /// </summary>
        /// <param name="generatedCode"></param>
        /// <param name="customTestId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpPost("upload/attach/custom/tests/{generatedCode}/{customTestId}/{fileReferenceId}")]
        public async Task AttachExitingFileToCustomTest(string generatedCode, int customTestId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .Include(p => p.CustomTests)
                .FirstOrDefaultAsync(p =>
                    p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode);

            if (oldExerciseReleaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }

            var exerciseTuple =
                    await _context.Exercises
                        .Where(p => p.Id == oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.ExerciseId)
                        .Select(p => new
                        {
                            p.CanUserCreateCustomTests,
                            p.IsPermanentlyLocked,
                            p.IsOnlyVisibleToOwner,
                            p.UserId
                        })
                        .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (exerciseTuple.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (exerciseTuple.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exerciseTuple.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            if (exerciseTuple.CanUserCreateCustomTests == false)
            {
                //user is not allowed to create custom tests

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.InvalidRequest, "not allowed to create custom test")));
                return;
            }

            //check if available working time is up
            //do not allow adding/changing custom tests after release finished (currently)
            //because we would invalidate stored custom test results
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }


            var oldTest = oldExerciseReleaseWithUserAsParticipation.CustomTests
                .FirstOrDefault(p => p.Id == customTestId
                                     && p.ExerciseReleaseId ==
                                     oldExerciseReleaseWithUserAsParticipation.ExerciseReleaseId);

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            var fileRef = await _context.FileReferenceUserFileAssets
                .FirstOrDefaultAsync(p => p.Id == fileReferenceId);

            if (fileRef == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "asset file not found")));
                return;
            }

            var connection = new CustomTestWithFileAsAssetReference()
            {
                FileReferenceUserFileAsset = fileRef,
                CustomTest = oldTest
            };

            try
            {
                _context.CustomTestWithFileAsAssetReferences.Add(connection);
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }

            //we need to send some information so the frontend can display the asset file

            var filePreview = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes,
                MimeType = fileRef.MimeType
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "", filePreview)));
        }

        ///  <summary>
        ///  removes the file from the custom test 
        /// 
        ///  IF the file was only attached to this last test then the file gets deleted
        ///    (because there are no other connection to this file)
        ///  </summary>
        ///  <param name="generatedCode"></param>
        /// <param name="customTestId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpDelete("upload/detach/custom/tests/{generatedCode}/{customTestId}/{fileReferenceId}")]
        public async Task DetachOrDeleteFileFromCustomTest(string generatedCode, int customTestId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            ExerciseReleaseWithUserAsParticipation oldExerciseReleaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .Include(p => p.CustomTests)
                .FirstOrDefaultAsync(p =>
                    p.UserId == userId && p.ExerciseRelease.GeneratedCode == generatedCode);

            if (oldExerciseReleaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "participation not found")));
                return;
            }

            var exerciseTuple =
                    await _context.Exercises
                        .Where(p => p.Id == oldExerciseReleaseWithUserAsParticipation.ExerciseRelease.ExerciseId)
                        .Select(p => new
                            {p.CanUserCreateCustomTests, p.IsPermanentlyLocked, p.IsOnlyVisibleToOwner, p.UserId})
                        .FirstOrDefaultAsync()
                ;

            if (exerciseTuple == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (exerciseTuple.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "exercise is permanently locked")));
                return;
            }

            if (exerciseTuple.IsOnlyVisibleToOwner)
            {
                //then only the owner can access the exercise
                if (userId != exerciseTuple.UserId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "exercise is only accessible by owner")));
                    return;
                }
            }

            if (exerciseTuple.CanUserCreateCustomTests == false)
            {
                //user is not allowed to create custom tests

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.InvalidRequest, "not allowed to create custom test")));
                return;
            }

            //check if available working time is up
            //do not allow adding/changing custom tests after release finished (currently)
            //because we would invalidate stored custom test results
            if (ReleaseHelper.IsTimeUp(oldExerciseReleaseWithUserAsParticipation,
                oldExerciseReleaseWithUserAsParticipation.ExerciseRelease, this))
            {
                return;
            }


            var oldTest = oldExerciseReleaseWithUserAsParticipation.CustomTests
                .FirstOrDefault(p => p.Id == customTestId
                                     && p.ExerciseReleaseId ==
                                     oldExerciseReleaseWithUserAsParticipation.ExerciseReleaseId);

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            var assetFileReference = await _context.CustomTestWithFileAsAssetReferences.FirstOrDefaultAsync(p =>
                p.FileReferenceUserFileAssetId == fileReferenceId && p.CustomTestId == oldTest.Id);

            if (assetFileReference == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference connection not found")));
                return;
            }

            //then check if this was the last reference to this file from an exercise...
            //for user files we need to check:
            //  custom test assets
            //  custom project description assets
            //  custom project test assets
            //only if the sum is 1 we can delete it!

            var referenceCount = await _context.CustomProjectTestWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            var referenceCount2 = await _context.CustomProjectDescriptionWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            var referenceCount3 = await _context.CustomTestWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            if (referenceCount + referenceCount2 + referenceCount3 == 1) //this is the last reference
            {
                //no test has a reference to this file... delete


                var fileReference = await _context.FileReferenceUserFileAssets
                    .FirstOrDefaultAsync(p => p.Id == assetFileReference.FileReferenceUserFileAssetId);


                if (fileReference == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference not found")));
                    return;
                }

                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        _context.FileReferenceUserFileAssets.Remove(fileReference);
                        await _context.SaveChangesAsync();

                        var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

                        Files.DeleteUploadedFile(targetHardDrivePath, fileReference.Id);

                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            else
            {
                try
                {
                    _context.CustomTestWithFileAsAssetReferences.Remove(assetFileReference);
                    await _context.SaveChangesAsync();
                }
                catch (Exception e)
                {
                    await base.HandleDbError(e);
                    return;
                }
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        /// <summary>
        /// returns all possible uploaded file ids the user can use
        /// </summary>
        /// <returns></returns>
        [HttpGet("upload/get/custom")]
        public async Task GetPossibleCustomFiles()
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            //user has only access to own asset files...

            var customTestFileReferences = await _context.CustomTestWithFileAsAssetReferences
                    .Include(p => p.FileReferenceUserFileAsset)
                    .Where(p => p.CustomTest.UserId == userId)
                    .ToListAsync()
                ;

            var customProjectDescriptionFileReferences = await _context
                    .CustomProjectDescriptionWithFileAsAssetReferences
                    .Include(p => p.FileReferenceUserFileAsset)
                    .Where(p => p.CustomProjectDescription.CustomProject.UserId == userId)
                    .ToListAsync()
                ;

            var customProjectCustomTestFileReferences = await _context.CustomProjectTestWithFileAsAssetReferences
                    .Include(p => p.FileReferenceUserFileAsset)
                    .Where(p => p.CustomProjectTest.CustomProject.UserId == userId)
                    .ToListAsync()
                ;


            var allFileReferencesForFrontend = new List<FilePreviewFromBackend>();

            customTestFileReferences =
                customTestFileReferences.DistinctBy(p => p.FileReferenceUserFileAssetId).ToList();

            foreach (var fileReference in customTestFileReferences)
            {
                allFileReferencesForFrontend.Add(new FilePreviewFromBackend()
                {
                    Id = fileReference.FileReferenceUserFileAsset.Id,
                    SizeInBytes = fileReference.FileReferenceUserFileAsset.SizeInBytes,
                    OriginalName = fileReference.FileReferenceUserFileAsset.OriginalName,
                    MimeType = fileReference.FileReferenceUserFileAsset.MimeType
                });
            }

            customProjectDescriptionFileReferences =
                customProjectDescriptionFileReferences.DistinctBy(p => p.FileReferenceUserFileAssetId).ToList();

            foreach (var fileReference in customProjectDescriptionFileReferences)
            {
                allFileReferencesForFrontend.Add(new FilePreviewFromBackend()
                {
                    Id = fileReference.FileReferenceUserFileAsset.Id,
                    SizeInBytes = fileReference.FileReferenceUserFileAsset.SizeInBytes,
                    OriginalName = fileReference.FileReferenceUserFileAsset.OriginalName,
                    MimeType = fileReference.FileReferenceUserFileAsset.MimeType
                });
            }

            customProjectCustomTestFileReferences =
                customProjectCustomTestFileReferences.DistinctBy(p => p.FileReferenceUserFileAssetId).ToList();

            foreach (var fileReference in customProjectCustomTestFileReferences)
            {
                allFileReferencesForFrontend.Add(new FilePreviewFromBackend()
                {
                    Id = fileReference.FileReferenceUserFileAsset.Id,
                    SizeInBytes = fileReference.FileReferenceUserFileAsset.SizeInBytes,
                    OriginalName = fileReference.FileReferenceUserFileAsset.OriginalName,
                    MimeType = fileReference.FileReferenceUserFileAsset.MimeType
                });
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<FilePreviewFromBackend>>(ResponseCode.Ok, "",
                        allFileReferencesForFrontend)));
        }

        #endregion
    }

    #region types

    #region fullbase

    /// <summary>
    /// a solution from the user for an exercise
    /// </summary>
    public class SolutionDoExerciseFullBase
    {
        public int UserId { get; set; }
        public int ReleaseId { get; set; }

        /// <summary>
        /// set through the release participation (not in the version)
        /// true: the student wants that his/her solution is not counted (for whatever), false: not
        /// </summary>
        public bool ShouldNotCount { get; set; }


        /// <summary>
        /// the main file id (where the entry point is declared, e.g. public static void main...)
        /// </summary>
        public int MainFileId { get; set; }

        /// <summary>
        /// the plang the solution version is for
        /// </summary>
        public int PLangId { get; set; }

        /// <summary>
        /// a note about the solution version
        /// </summary>
        public string Note { get; set; }

        /// <summary>
        /// the datetime when the version was created
        /// </summary>
        public DateTime LastUpdatedAt { get; set; }

        /// <summary>
        /// the files for this solution version
        /// </summary>
        public List<SolutionFileDoExerciseFullBase> SolutionFiles { get; set; }

        /// <summary>
        /// true: load if existing, false: not or the normal solution is still active (editable), time is not used up
        /// only important/set on initial site load (not when resetting the user solution)
        /// </summary>
        public bool ShouldLoadAfterSolution { get; set; }

        /// <summary>
        /// true: user already created one, false: not
        /// only important/set on initial site load (not when resetting the user solution)
        /// </summary>
        public bool HasLoadAfterSolution { get; set; }

        public SolutionDoExerciseFullBase()
        {
            SolutionFiles = new List<SolutionFileDoExerciseFullBase>();
        }
    }

    /// <summary>
    /// a solution file (part of the solution verion)
    /// </summary>
    public class SolutionFileDoExerciseFullBase : ExerciseCodeTemplateFileFullBase
    {
        /// <summary>
        /// ui property, true: is displayed as tab, false: not (only displayed in project explorer)
        /// </summary>
        public bool UiIsDisplayed { get; set; }

        /// <summary>
        /// null or the template file id from which the file was created
        /// </summary>
        public int? TemplateFileId { get; set; }
    }

    #endregion

    #region different types for front and backend

    /// <summary>
    /// backend cannot receive a array/list (json array) so this is a wrapper
    /// </summary>
    public class UpdateCustomTestIndicesTupleWrapper
    {
        public List<UpdateCustomTestIndicesTuple> Tuples { get; set; }
    }

    /// <summary>
    /// a tuple with the test id and the new display index
    /// used when the user changes the display index of one or all custom tests
    /// </summary>
    public class UpdateCustomTestIndicesTuple
    {
        /// <summary>
        /// the custom test id
        /// </summary>
        public int Id { get; set; }

        /// <summary>
        /// the new display index for this test
        /// </summary>
        public int NewDisplayIndex { get; set; }
    }

    /// <summary>
    /// a solution tuple for the backend (includes one solution version and a list of all known solutions versions from the user)
    /// </summary>
    public class SolutionTupleFromBackend
    {
        /// <summary>
        /// the solution version for the frontend
        /// </summary>
        public SolutionDoExerciseFullBase Solution { get; set; }

        /// <summary>
        /// for every solution file in the new backend version the old id
        /// </summary>
        public List<FileIdAssociation> FileIdAssociationMap { get; set; }
    }

    /// <summary>
    /// if we change/create a new file in the frontend then we send back the saved version (backend version) to the frontend
    /// because the frontend can replace the file (else the undo history would be lost/corrupted) it needs to set the
    /// file content directly. for this we need to know which old/frontend file had which content
    /// </summary>
    public class FileIdAssociation
    {
        /// <summary>
        /// the file id at frontend
        /// </summary>
        public int FrontendFileId { get; set; }

        /// <summary>
        /// the backend id for the same file
        /// </summary>
        public int BackendFileId { get; set; }
    }


    /// <summary>
    /// a preview of a solution version (includes just some basic data so that the user knows about the version (just the id is not so informative))
    /// </summary>
    public class SolutionVersionPreview
    {
        /// <summary>
        /// the id of the solution version
        /// </summary>
        public int Id { get; set; }

        /// <summary>
        /// a note about the version
        /// </summary>
        public string Note { get; set; }

        /// <summary>
        /// when the version was created
        /// </summary>
        public DateTime CreatedAt { get; set; }
    }


    /// <summary>
    /// the exercise for the do exercise page
    /// frontend type: DoExerciseBase
    /// </summary>
    public class DoExerciseFromBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the short description
        /// </summary>
        public string ShortDescription { get; set; }

        /// <summary>
        /// true: user can create own tests, false: not
        /// </summary>
        public bool CanUserCreateCustomTests { get; set; }

        /// <summary>
        /// true: user can create own files (tabs), false: not
        /// </summary>
        public bool CanUserCreateFiles { get; set; }

        /// <summary>
        /// the last edited plang to pre select first when entering the do exercise page
        /// </summary>
        public int LastEditedPLangId { get; set; }

        /// <summary>
        /// exercise and all dependent entities should be readonly (no edit/no delete)! (recursively / transitively) 
        /// </summary>
        public bool IsPermanentlyLocked { get; set; }

        /// <summary>
        /// the code templates for every plang (used to reset the files)
        /// </summary>
        public List<ExerciseCodeTemplateFullBase> CodeTemplates { get; set; }

        /// <summary>
        /// the exercise description
        /// </summary>
        public ExerciseDescriptionFullBase ExerciseDescription { get; set; }

        /// <summary>
        /// the tests (without results)
        /// </summary>
        public List<ExerciseTestFromBackend> Tests { get; set; }

        /// <summary>
        /// the custom tests (without results)
        /// </summary>
        public List<CustomTestFromBackend> CustomTests { get; set; }

        /// <summary>
        /// the settings for all custom tests created for this exercise
        /// </summary>
        public TestSettingsFullBase DefaultCustomTestSettings { get; set; }

        /// <summary>
        /// the related exercise release
        /// </summary>
        public ExerciseReleaseDoExerciseFromBackend ExerciseRelease { get; set; }

        /// <summary>
        /// the date time when the user has accessed the exercise for the first time (not actually accessed but could have accessed e.g. wehn he/she enters the release code)
        /// used to calculate the left working time...
        /// </summary>
        public DateTime FirstAccessedAt { get; set; }

        /// <summary>
        /// the lock solutions flag from the participation
        /// </summary>
        public bool LockSolutionsFlag { get; set; }

        public DoExerciseFromBackend()
        {
            CodeTemplates = new List<ExerciseCodeTemplateFullBase>();
            Tests = new List<ExerciseTestFromBackend>();
            CustomTests = new List<CustomTestFromBackend>();
        }
    }

    #region stored test results

    /// <summary>
    /// the test results (ever plang has own test results for the same tests)
    /// </summary>
    public class StoredTestResultTupleFromBackend
    {
        /// <summary>
        /// the normal test results
        /// </summary>
        public List<StoredTestResultFromBackend> TestResults { get; set; }

        /// <summary>
        /// the results of the custom tests
        /// </summary>
        public List<StoredTestResultFromBackend> CustomTestResults { get; set; }
    }

    /// <summary>
    /// the submission result
    /// </summary>
    public class AssessmentForUserWithTestResults
    {
        public AssessmentResultForUser AssessmentResult { get; set; }
        public List<AssessedSubmitTestFromBackend> SubmitTestResults { get; set; }
    }

    public class AssessmentResultForUser
    {
        /// <summary>
        /// the used plang
        /// </summary>
        public int PLangId { get; set; }

        public int ReleaseId { get; set; }
        public int UserId { get; set; }

        /// <summary>
        /// the summed test points from the submit tests
        /// </summary>
        public int? SubmitTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable submit test points 
        /// </summary>
        public int? MaxSubmitTestPoints { get; set; }

        /// <summary>
        /// the summed points from the normal test
        /// </summary>
        public int? NormalTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable normal test points 
        /// </summary>
        public int? MaxNormalTestPoints { get; set; }

        /// <summary>
        /// optional the manual given poitns (by the tutor)
        /// </summary>
        public int? ManualPoints { get; set; }

        /// <summary>
        /// the note for the student why he/she gets the manual points (feedback)
        /// </summary>
        public string FeedbackForStudent { get; set; }
    }

    /// <summary>
    /// a submit test with result
    /// </summary>
    public class AssessedSubmitTestFromBackend : ExerciseTestFromBackend
    {
        /// <summary>
        /// the submit test result
        /// </summary>
        public StoredTestResultFromBackend Result { get; set; }
    }

    /// <summary>
    /// this is not a real answer from the test server just the stored result...
    /// </summary>
    public class StoredTestResultFromBackend
    {
        /// <summary>
        /// the id of the test (custom or normal)
        /// </summary>
        public int TestId { get; set; }

        /// <summary>
        /// null: not ran yet, true: test was successful, false: not
        /// </summary>
        public bool? Passed { get; set; }

        /// <summary>
        /// null: not ran yet, true: code compiled, false: not
        /// </summary>
        public bool? HasCompiled { get; set; }

        /// <summary>
        /// null: not ran yet, the exit code of the user program, not used yet (not event set)
        /// </summary>
        public int? ProgramExitCode { get; set; }

        /// <summary>
        /// null: not ran yet, the test result code (see test server codes)
        /// </summary>
        public int? TestResultCode { get; set; }

        /// <summary>
        /// the conversation protocol between the user program and the test runner
        /// </summary>
        public List<string> Protocol { get; set; }


        /// <summary>
        /// the test runner version
        /// can be null
        /// </summary>
        public string RunnerVersion { get; set; }

        /// <summary>
        /// time for compiling in ms
        /// can be null if compiling was not finished or done
        /// </summary>
        public int? TimeForCompiling { get; set; }

        /// <summary>
        /// time needed for execution in ms
        /// can be null if execution was not finished or done
        /// </summary>
        public int? TimeForUserProgram { get; set; }

        /// <summary>
        /// null: not ran (e.g. because of test server error)
        /// else: the timeout in ms used for test execution
        /// </summary>
        public int? TimeoutInMsUsed { get; set; }

        /// <summary>
        /// null: not ran (e.g. because of test server error)
        /// else: the timeout in ms used for compiling only
        /// </summary>
        public int? CompileTimeoutInMsUsed { get; set; }

        /// <summary>
        /// the test server version
        /// can be null
        /// </summary>
        public string TestServerVersion { get; set; }
    }

    #endregion

    #region custom tests

    /// <summary>
    /// a custom test
    /// </summary>
    public class CustomTestFromBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the content
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// the index for the sorting
        /// </summary>
        public int DisplayIndex { get; set; }

        /// <summary>
        /// the test case type id
        /// </summary>
        public int TestTypeId { get; set; }

        /// <summary>
        /// the test assets (files)
        /// </summary>
        public List<FilePreviewFromBackend> Files { get; set; }
    }


    /// <summary>
    /// a custom test (IGNORE TestCaseSettings because we use the defaultTestCase settings from the exercise)
    /// </summary>
    public class CustomTestForBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the content
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// the index for the sorting
        /// </summary>
        public int DisplayIndex { get; set; }

        /// <summary>
        /// the test case type id
        /// </summary>
        public int TestTypeId { get; set; }

        /// <summary>
        /// the test assets (files)
        /// this is not used but the frontend type uses this so we keep this here...
        /// </summary>
        public List<FilePreviewFromBackend> Files { get; set; }
    }

    #endregion

    /// <summary>
    /// the connected (to the exercise) exercise release, we need this because the working time 
    /// could be limited
    /// </summary>
    public class ExerciseReleaseDoExerciseFromBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// true: exercise is still released,
        /// false: not,
        /// when the release ends the user cannot modify the solution but look at it (and the result of the submit test if any)
        /// </summary>
        public bool IsReleased { get; set; }

        /// <summary>
        /// the release code
        /// </summary>
        public string ReleaseCode { get; set; }

        /// <summary>
        /// null if never ends or date time when the release ends (release could end before the AvailableWorkingTimeInMinutes is reached
        /// </summary>
        public DateTime? AutomaticEndAt { get; set; }

        /// <summary>
        /// the available working time for the exercise in minutes
        /// </summary>
        public int AvailableWorkingTimeInMinutes { get; set; }

        /// <summary>
        /// true: the working time is limited, false: ignore the AvailableWorkingTimeInMinutes
        /// </summary>
        public bool HasLimitedWorkingTime { get; set; }

        /// <summary>
        /// the release duration type (to know if we can ignore the AutomaticEndAt datetime)
        /// </summary>
        public int ReleaseDurationType { get; set; }

        /// <summary>
        /// the release start type (to know if we can trust automatic end at)
        /// </summary>
        public int ReleaseStartType { get; set; }

        /// <summary>
        /// the p lang for the release
        /// </summary>
        public int PLangId { get; set; }

        /// <summary>
        /// true: assessment is finished, false: not
        /// </summary>
        public bool HasAutomaticAssessmentFinished { get; set; }


        /// <summary>
        /// e.g. in an electronic exam (e-exam) we might have multiple windows in e.g. the safe exam browser
        /// the student could copy & past exercises from other systems into yapex and try to compile, execute code...
        ///  
        /// true: try to clear the clipboard if the tab is changed...
        ///     show an overlay every time the student tabs or clicks outside of the browser window
        ///     and when he/she wants to resume he/she must click a button to proceed
        /// false: normal
        /// 
        /// frontend only option...
        /// </summary>
        public bool ShouldClearClipboard { get; set; }

        /// <summary>
        /// in the do exercise site the link mode always overwrites this setting!
        ///
        /// true: hide the site header bar (where the menu, user menu is located)
        ///    so that a student cannot navigate away from the do exercise site (e.g. in an exam)
        ///
        /// false: not hide the bar
        /// </summary>
        public bool HideSiteHeaderBar { get; set; }

        /// <summary>
        /// hides all print options (e.g. print task description, code editor) for the exercise
        /// </summary>
        public bool HidePrintOptions { get; set; }

        /// <summary>
        /// hides all actions and buttons the user could use to leave the exercise (and e.g. select another exercise).
        /// NOTE The buttons in the site header bar are not affected (use the appropriate option for that).
        /// <see cref="HideSiteHeaderBar"/>
        /// </summary>
        public bool HideExerciseLeaveActions { get; set; }

        /// <summary>
        /// some options might hide the user header bar, this will display an additional button to show (and download) the log
        /// </summary>
        public bool ShowAdditionalLogButton { get; set; }

        /// <summary>
        /// the max points for manual assessment
        /// NULL for no maximum
        ///
        /// however, one can give more points than this (e.g. extra points)
        /// or if this null manual points are still possible to give
        /// </summary>
        public int? MaxManualPoints { get; set; }
    }

    #endregion

    #endregion
}
